@@ -1,3 +1,4 @@
+import logging
 import numpy as np
 import xarray as xr
 from pydantic import BaseModel, Extra, validator
@@ -5,6 +6,7 @@ from typing import Any, Dict, List, Optional, Union
 from numpy.typing import NDArray
 from .base import QualityChecker
 
+
 __all__ = [
     "CheckMissing",
     "CheckMonotonic",
@@ -25,6 +27,8 @@ __all__ = [
     "CheckWarnDelta",
 ]
 
+logger = logging.getLogger(__name__)
+
 
 class CheckMissing(QualityChecker):
     """---------------------------------------------------------------------------------
@@ -77,8 +81,18 @@ class CheckMonotonic(QualityChecker):
     parameters: Parameters = Parameters()
 
     def run(self, dataset: xr.Dataset, variable_name: str) -> NDArray[np.bool8]:
+
         variable = dataset[variable_name]
 
+        if variable.values.dtype.kind in {"U", "S"}:  # type: ignore
+            logger.warning(
+                "Variable '%s' has dtype '%s', which is currently not supported for"
+                " monotonicity checks.",
+                variable_name,
+                variable.values.dtype,  # type: ignore
+            )
+            return np.full(variable.shape, False, dtype=np.bool8)
+
         axis = self.get_axis(variable)
         diff: NDArray[Any] = np.diff(variable.data, axis=axis)  # type: ignore
 
@@ -350,7 +364,8 @@ class _CheckDelta(_ThresholdChecker):
         data: NDArray[Any] = var_data.data
         axis = var_data.get_axis_num(self.parameters.dim)
 
-        diff: NDArray[Any] = np.absolute(np.diff(data, axis=axis, prepend=data[0]))  # type: ignore
+        prepend = np.expand_dims(np.take(data, 0, axis=axis), axis=axis)  # type: ignore
+        diff: NDArray[Any] = np.absolute(np.diff(data, axis=axis, prepend=prepend))  # type: ignore
         failures = diff > threshold if self.allow_equal else diff >= threshold
 
         return failures
